{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# constant embedding 分区"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "from testing import viz_expr # 可视化 relay"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "import tvm\n",
    "from tvm import relay\n",
    "from tvm.relay.dataflow_pattern import *\n",
    "\n",
    "def conv_bias_relu(x, w, b):\n",
    "    conv2d = relay.op.nn.conv2d(x, w)\n",
    "    bias_add = relay.op.nn.bias_add(conv2d, b)\n",
    "    relu = relay.op.nn.relu(bias_add)\n",
    "    return relu"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "构建计算图："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "x = relay.var(\"x\")\n",
    "w = relay.var(\"w\")\n",
    "wc = relay.const(1)\n",
    "b = relay.var(\"b\")\n",
    "\n",
    "relu = conv_bias_relu(x, w, b)\n",
    "reluc = conv_bias_relu(x, wc, b)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "# viz_expr(relu)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %w, %b) {\n",
      "  %0 = nn.conv2d(%x, %w, padding=[0, 0, 0, 0]);\n",
      "  %1 = nn.bias_add(%0, %b);\n",
      "  nn.relu(%1)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "print(tvm.IRModule.from_expr(relu))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %b) {\n",
      "  %0 = nn.conv2d(%x, 1, padding=[0, 0, 0, 0]);\n",
      "  %1 = nn.bias_add(%0, %b);\n",
      "  nn.relu(%1)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "print(tvm.IRModule.from_expr(reluc))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "构建模式(解除 `wildcard()` 匹配)："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 7,
   "metadata": {},
   "outputs": [],
   "source": [
    "pattern = is_op(\"nn.relu\")(\n",
    "    is_op(\"nn.bias_add\")(is_op(\"nn.conv2d\")(wildcard(), wildcard()), wildcard())\n",
    ")"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "分割计算图："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %w, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, %FunctionVar_0_2, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, %FunctionVar_0_1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_2);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, %w, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(relu)\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, %FunctionVar_0_2, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, %FunctionVar_0_1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_2);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, 1, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(reluc)\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "构建模式(解除 input 匹配)："
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 10,
   "metadata": {},
   "outputs": [],
   "source": [
    "pattern = is_op(\"nn.relu\")(\n",
    "    is_op(\"nn.bias_add\")(is_op(\"nn.conv2d\")(wildcard(), is_var()), wildcard())\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 11,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %w, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, %FunctionVar_0_2, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, %FunctionVar_0_1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_2);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, %w, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(relu)\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "常量不是输入:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 12,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %b) {\n",
      "  %0 = nn.conv2d(%x, 1, padding=[0, 0, 0, 0]);\n",
      "  %1 = nn.bias_add(%0, %b);\n",
      "  nn.relu(%1)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(reluc)\n",
    "print(tvm.IRModule.from_expr(partitioned)) "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "检查常量匹配的嵌入:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 13,
   "metadata": {},
   "outputs": [],
   "source": [
    "pattern = is_op(\"nn.relu\")(\n",
    "    is_op(\"nn.bias_add\")(is_op(\"nn.conv2d\")(wildcard(), is_constant()), wildcard())\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 14,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %w, %b) {\n",
      "  %0 = nn.conv2d(%x, %w, padding=[0, 0, 0, 0]);\n",
      "  %1 = nn.bias_add(%0, %b);\n",
      "  nn.relu(%1)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(relu)\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 15,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, 1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_1);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(reluc)\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "检查常量 ExprPatterns 的嵌入:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 16,
   "metadata": {},
   "outputs": [],
   "source": [
    "pattern = is_op(\"nn.relu\")(\n",
    "    is_op(\"nn.bias_add\")(is_op(\"nn.conv2d\")(wildcard(), is_expr(wc)), wildcard())\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 17,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %w, %b) {\n",
      "  %0 = nn.conv2d(%x, %w, padding=[0, 0, 0, 0]);\n",
      "  %1 = nn.bias_add(%0, %b);\n",
      "  nn.relu(%1)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(relu)\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 18,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, 1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_1);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(reluc)\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "检查 Alt 匹配的提升/嵌入:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 19,
   "metadata": {},
   "outputs": [],
   "source": [
    "pattern = is_op(\"nn.relu\")(\n",
    "    is_op(\"nn.bias_add\")(is_op(\"nn.conv2d\")(wildcard(), is_var() | is_constant()), wildcard())\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 20,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %w, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, %FunctionVar_0_2, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, %FunctionVar_0_1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_2);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, %w, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(relu) # lifted\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 21,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, 1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_1);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(reluc) # embeded\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "检查 Alt 匹配的提升/嵌入是否与其他排序一致:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 22,
   "metadata": {},
   "outputs": [],
   "source": [
    "pattern = is_op(\"nn.relu\")(\n",
    "    is_op(\"nn.bias_add\")(is_op(\"nn.conv2d\")(wildcard(), is_constant() | is_var()), wildcard())\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %w, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, %FunctionVar_0_2, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, %FunctionVar_0_1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_2);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, %w, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(relu) # lifted\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "def @main(%x, %b) {\n",
      "  %2 = fn (%FunctionVar_0_0, %FunctionVar_0_1, PartitionedFromPattern=\"nn.conv2d_nn.bias_add_nn.relu_\") {\n",
      "    %0 = nn.conv2d(%FunctionVar_0_0, 1, padding=[0, 0, 0, 0]);\n",
      "    %1 = nn.bias_add(%0, %FunctionVar_0_1);\n",
      "    nn.relu(%1)\n",
      "  };\n",
      "  %2(%x, %b)\n",
      "}\n",
      "\n"
     ]
    }
   ],
   "source": [
    "partitioned = pattern.partition(reluc) # embeded\n",
    "print(tvm.IRModule.from_expr(partitioned))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "py312x",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.12.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
